@@ -72,6 +72,15 @@
 #ifndef NO_NAMESPACE_GIAC
 namespace giac {
 #endif // ndef NO_NAMESPACE_GIAC
+  void wait_1ms(int ms=1);
+  template<class U>
+  inline unsigned sizeinbase2(U n){
+    unsigned i=0;
+    for (;n;++i){
+      n >>= 1;
+    }
+    return i;
+  }
   std::vector<int> operator % (const std::vector<int> & a,int modulo);
   std::vector<int> operator / (const std::vector<int> & v,const std::vector<int> & b);
   std::vector<int> operator % (const std::vector<int> & v,const std::vector<int> & b);
@@ -479,6 +488,116 @@ namespace giac {
     unsigned int _capacity;
   };
 
+#if 0
+  template<class T>
+  class vector_size32 {
+  public:
+    struct {
+      unsigned int taille;
+      T v0;
+      T v1;
+      T v2;
+      T v3;
+      T v4;
+      T v5;
+      T v6;
+    };
+    vector_size32(){ taille=1;}
+    vector_size32(size_t t){ 
+      if (t>7){
+	mytab<T> * mytabptr = (mytab<T> *) this;
+	mytabptr->tab = new T[t];
+	mytabptr->_size = mytabptr->_capacity = t;
+	int * ptr=(int *)mytabptr->tab;
+	for (size_t i=0;i<t;++ptr,++i)
+	  *ptr=0;
+      }
+      else { 
+	taille=2*t+1; 
+	* (int *) &v0=0; * (longlong *) &v1=0;
+	* (longlong *) &v3=0;
+	* (longlong *) &v5=0;
+      }
+    }
+    vector_size32(size_t t,const T & elem){ 
+      if (t>7){ 
+	mytab<T> * mytabptr = (mytab<T> *) this;
+	mytabptr->tab = new T[t];
+	mytabptr->_size = mytabptr->_capacity = t;
+	T * ptr=mytabptr->tab;
+	for (size_t i=0;i<t;++ptr,++i)
+	  *ptr=elem;
+      }
+      else { 
+	taille=2*t+1; 
+	v6=v5=v4=v3=v2=v1=v0=elem;
+      }
+    }
+    ~vector_size32(){
+      if (taille%2==0){
+	mytab<T> * mytabptr = (mytab<T> *) this;
+	delete [] mytabptr->tab;
+      }
+    }
+    void push_back(T t){
+      if (taille %2){
+	if (taille==2*7+1){
+	  T * ptr = new T[16];
+	  *ptr=v0; ++ptr;
+	  *ptr=v1; ++ptr;
+	  *ptr=v2; ++ptr;
+	  *ptr=v3; ++ptr;
+	  *ptr=v4; ++ptr;
+	  *ptr=v5; ++ptr;
+	  *ptr=v6; ++ptr;
+	  *ptr=t;
+	  mytab<T> * mytabptr = (mytab<T> *) this;
+	  mytabptr->tab = ptr-7;
+	  mytabptr->_size = 8;
+	  mytabptr->_capacity = 16;
+	  return;
+	}
+	*(&v0+taille/2)=t;
+	taille += 2;
+      }
+      else {
+	mytab<T> * mytabptr = (mytab<T> *) this;
+	if (mytabptr->_size>=mytabptr->_capacity){
+	  mytabptr->_capacity *=2;
+	  T * ptr = new T[mytabptr->_capacity];
+	  memcpy(ptr,mytabptr->tab,mytabptr->_size*sizeof(T));
+	  delete [] mytabptr->tab;
+	  mytabptr->tab=ptr;
+	}
+	mytabptr->tab[mytabptr->_size]=t;
+	++mytabptr->_size;
+      }
+    }
+    void pop_back(){
+      if (taille%2)
+	taille -=2;
+      else {
+	mytab<T> * mytabptr = (mytab<T> *) this;
+	--mytabptr->_size;
+      }
+    }
+    void clear(){
+      if (taille%2)
+	taille=1;
+      else {
+	mytab<T> * mytabptr = (mytab<T> *) this;
+	mytabptr->_size=0;
+      }
+    }
+    T operator [] (size_t pos) const {
+      if (taille%2)
+	return *(&v0+pos);
+      else {
+	mytab<T> * mytabptr = (mytab<T> *) this;
+	return mytabptr->tab[pos];
+      }
+    }
+#else
   template<class T>
   class vector_size32 {
   public:
@@ -531,12 +650,8 @@ namespace giac {
 	  mytabptr->_capacity = 6;
 	  return;
 	}
+	*(&v0+taille/2)=t;
 	taille += 2;
-	if (taille==7)
-	  v2=t;
-	else { 
-	  if (taille==5) v1=t; else v0=t; 
-	}
       }
       else {
 	mytab<T> * mytabptr = (mytab<T> *) this;
@@ -569,12 +684,13 @@ namespace giac {
     }
     T operator [] (size_t pos) const {
       if (taille%2)
-	return pos?((pos==2)?v2:v1):v0;
+	return *(&v0+pos); // pos?((pos==2)?v2:v1):v0;
       else {
 	mytab<T> * mytabptr = (mytab<T> *) this;
 	return mytabptr->tab[pos];
       }
     }
+#endif
     typename std::vector<T>::const_iterator begin() const {
       if (taille %2)
 	return typename std::vector<T>::const_iterator(&v0);
@@ -1069,7 +1185,7 @@ namespace giac {
   }
 
   template<class T>
-  void in_out_heap(T * tab,unsigned size,T value){
+  void in_out_heap(T * tab,size_t size,T value){
     unsigned childindex=2,holeindex=0;
     while (childindex<size){
       // find largest child until end of tab
@@ -1210,7 +1326,9 @@ namespace giac {
 	// using heap of chains
 	// std::vector< vector_size64< std::pair<unsigned,unsigned> > > vindex(v1s);
 	std::vector< std::vector< std::pair<unsigned,unsigned> > > vindex(v1s);
+#ifdef HEAP_STATS
 	double count1=0,count2=0,total=double(v1s)*v2s;
+#endif
 	U_unsigned<U> * heap = new U_unsigned<U>[v1s] ; // pointers to v2 monomials
 	U_unsigned<U> * heap0, *heapbeg=heap,* heapend=heap+v1s;
 	for (it1=it1beg,heap0=heap;heap0!=heapend;++heap0,++it1){
@@ -1254,12 +1372,21 @@ namespace giac {
 	  // erase top node, then push each element of the incremented top chain 
 	  it=nouveau.begin();
 	  itend=nouveau.end();
+	  U prevu=0; int previndex=-1;
 	  for (;it!=itend;++it){
 	    u=(it1beg+it->first)->u+(it2beg+it->second)->u;
+	    if (u==prevu && previndex>=0){
+	      vindex[previndex].push_back(*it);
+#ifdef HEAP_STATS
+	      ++count1;
+#endif
+	      continue;
+	    }
+	    prevu=u;
 	    // check if u is in the path to the root of the heap
 	    unsigned holeindex=unsigned(heapend-heapbeg),parentindex;
 	    if (holeindex && u==heapbeg->u){
-	      vindex[heapbeg->v].push_back(*it);
+	      vindex[previndex=heapbeg->v].push_back(*it);
 #ifdef HEAP_STATS
 	      ++count1;
 #endif
@@ -1276,15 +1403,15 @@ namespace giac {
 #ifdef HEAP_STATS
 		++count2;
 #endif
-		vindex[(heapbeg+parentindex)->v].push_back(*it);
+		vindex[previndex=(heapbeg+parentindex)->v].push_back(*it);
 		break;
 	      }
 	      holeindex=parentindex;
 	    }
 	    if (!done){
 	      // not found, add a new node to the heap
-	      vindex[heapend->v].clear();
-	      vindex[heapend->v].push_back(*it);
+	      vindex[previndex=heapend->v].clear();
+	      vindex[previndex].push_back(*it);
 	      heapend->u=u;
 	      ++heapend;
 #ifdef USTL
@@ -1711,6 +1838,7 @@ namespace giac {
 	  // erase top node, then push each element of the incremented top chain 
 	  it=nouveau.begin();
 	  itend=nouveau.end();
+	  U prevu=0; int previndex=-1;
 	  for (;it!=itend;++it){
 	    it1=it1beg+it->first;
 	    u1=it1->u;
@@ -1718,10 +1846,16 @@ namespace giac {
 	    d2=v2deg-(d-d1);
 	    it2beg=(*v2ptrs)[d2];
 	    u=u1+(it2beg+it->second)->u;
+	    if (u==prevu && previndex>=0){
+	      vptr=&(*vsmallindexptr)[previndex];
+	      vptr->push_back(*it);
+	      continue;
+	    }
+	    prevu=u;
 	    // check if u is in the path to the root of the heap
 	    unsigned holeindex=heapend-heapbeg,parentindex;
 	    if (holeindex && u==heapbeg->u){
-	      vptr=&(*vsmallindexptr)[heapbeg->v];
+	      vptr=&(*vsmallindexptr)[previndex=heapbeg->v];
 	      vptr->push_back(*it);
 	      continue;
 	    }
@@ -1732,7 +1866,7 @@ namespace giac {
 		break;
 	      if (u==(heapbeg+parentindex)->u){
 		done=true;
-		vptr=&(*vsmallindexptr)[(heapbeg+parentindex)->v];
+		vptr=&(*vsmallindexptr)[previndex=(heapbeg+parentindex)->v];
 		vptr->push_back(*it);
 		break;
 	      }
@@ -1740,7 +1874,7 @@ namespace giac {
 	    }
 	    if (!done){
 	      // not found, add a new node to the heap
-	      vptr=&(*vsmallindexptr)[heapend->v];
+	      vptr=&(*vsmallindexptr)[previndex=heapend->v];
 	      vptr->clear();
 	      vptr->push_back(*it);
 	      heapend->u=u;
@@ -1800,6 +1934,7 @@ namespace giac {
 	  // erase top node, then push each element of the incremented top chain 
 	  it=nouveau.begin();
 	  itend=nouveau.end();
+	  U prevu=0; int previndex=-1;
 	  for (;it!=itend;++it){
 	    it1=it1beg+it->first;
 	    u1=it1->u;
@@ -1807,10 +1942,16 @@ namespace giac {
 	    d2=v2deg-(d-d1);
 	    it2beg=(*v2ptrs)[d2];
 	    u=u1+(it2beg+it->second)->u;
+	    if (u==prevu && previndex>=0){
+	      vptr=&(*vindexptr)[previndex];
+	      vptr->push_back(*it);
+	      continue;
+	    }
+	    prevu=u;
 	    // check if u is in the path to the root of the heap
 	    unsigned holeindex=heapend-heapbeg,parentindex;
 	    if (holeindex && u==heapbeg->u){
-	      vptr=&(*vindexptr)[heapbeg->v];
+	      vptr=&(*vindexptr)[previndex=heapbeg->v];
 	      vptr->push_back(*it);
 	      continue;
 	    }
@@ -1821,7 +1962,7 @@ namespace giac {
 		break;
 	      if (u==(heapbeg+parentindex)->u){
 		done=true;
-		vptr=&(*vindexptr)[(heapbeg+parentindex)->v];
+		vptr=&(*vindexptr)[previndex=(heapbeg+parentindex)->v];
 		vptr->push_back(*it);
 		break;
 	      }
@@ -1829,7 +1970,7 @@ namespace giac {
 	    }
 	    if (!done){
 	      // not found, add a new node to the heap
-	      vptr=&(*vindexptr)[heapend->v];
+	      vptr=&(*vindexptr)[previndex=heapend->v];
 	      vptr->clear();
 	      vptr->push_back(*it);
 	      heapend->u=u;
@@ -1944,7 +2085,9 @@ namespace giac {
     }
     // if array multiplication is faster, set prod
     bool use_heap = (heap_mult<0) || (heap_mult>0 && v1v2>heap_mult);
-    if (!prod && use_heap && nthreads<2) 
+    if (!prod && use_heap 
+	&& nthreads<2 // ???
+	) // multi-thread heap disabled because of locks by inserting in chains 
       return false;  
     if (debug_infolevel>20){
       CERR << "// " << CLOCK() << "using threaded " ;
@@ -2034,7 +2177,8 @@ namespace giac {
 	possible_size += arg[i].vptr->size();	
       }
       delete [] heapptr;
-      delete vindexptr;
+      if (vindexptr) delete vindexptr;
+      if (vsmallindexptr) delete vsmallindexptr;
     }
     else {
       std::vector<int> in_progress;
@@ -2087,7 +2231,7 @@ namespace giac {
 		arg[i].v1ptrs=&v1it;
 		arg[i].v2ptrs=&v2it;
 		arg[i].vptr = new std::vector< T_unsigned<T,U> >;
-		arg[i].vptr->reserve(arg[k].vptr->size());
+		arg[i].vptr->reserve((3*arg[k].vptr->size())/2); // ???
 		arg[i].degdiv=degdiv;
 		arg[i].current_deg=i;
 		arg[i].reduce=reduce;
@@ -2255,6 +2399,28 @@ namespace giac {
   inline bool hashdivrem_finish_later(const mpz_class & a){return true;}
 #endif
 
+  template<class U>
+  inline bool one_index_smaller(U u,U v,const std::vector<int> & varsshift){
+    if (u<v) 
+      return true;
+    return false;
+    /*
+    // the code below is too slow
+    std::vector<int>::const_iterator it=varsshift.begin(),itend=varsshift.end();
+    for (;it!=itend;++it){
+      int shift=*it;
+      U u1=(u>>shift);
+      U v1=(v>>shift);
+      if (u1<v1) 
+	return true;
+      u -= (u1 << shift);
+      v -= (v1 << shift);
+    }
+    return false;
+    */
+  }
+
+  // #define HEAP_STATS
   // note that U may be of type vector of int or an int
   // + is used to multiply monomials and - to divide
   // / should return the quotient of the main variable exponent
@@ -2333,7 +2499,7 @@ namespace giac {
 	&& heap_mult>=0 && a.front().u < 512e6/sizeof(T)){
       U umax=a.front().u,u;
       if (debug_infolevel>1)
-	CERR << "array division, a size " << a.size() << " b size " << b.size() << " u " << umax << std::endl;
+	CERR << CLOCK()*1e-6 << " array division, a size " << a.size() << " b size " << b.size() << " u " << umax << std::endl;
       // array division
       T * rem = new T[unsigned(umax+1)];
       for (u=0;u<=umax;++u)
@@ -2438,7 +2604,7 @@ namespace giac {
       }
       delete [] rem;
       return 1;
-    }
+    } // end array division
     bool use_heap=false && 
       (heap_mult>0 
        && v1v2>=heap_mult
@@ -2448,11 +2614,21 @@ namespace giac {
 	(quo_only && quo_only<3)){
 	  // quo_only<3){
       bool norem=quo_only==2 || quo_only==-2;
+      norem=false; // currently disabled, it's not clear it wins something
+      std::vector<int> varsshift; // vars=2^varsshift
+      if (norem){
+	for (size_t i=0;i<vars.size();++i){
+	  int j=sizeinbase2(vars[i])-1;
+	  if (vars[i]!=(U(1)<<j))
+	    norem=false;
+	  varsshift.push_back(j);
+	}
+      }
 #ifdef HEAP_STATS
-      unsigned chain=0,nochain=0,typeopreduce=0;
+      unsigned chain=0,nochain=0,typeopreduce=0,nullq=0;
 #endif
       if (debug_infolevel>1)
-	CERR << "heap division, a size " << a.size() << " b size " << b.size() << " vars " << vars << std::endl;
+	CERR << CLOCK()*1e-6 << " heap division, a size " << a.size() << " b size " << b.size() << " vars " << vars << std::endl;
       // heap division:
       // ita an iterator on a, initial value a.begin()
       // a heap with the current state of q*b, initialized to empty heap
@@ -2497,7 +2673,7 @@ namespace giac {
 	      std::vector< std::pair<unsigned,unsigned> >::iterator it,itend;
 	      nouveau.clear();
 	      while (heapend!=heapbeg && heapu==heapbeg->u){
-		nouveau.clear();
+		//nouveau.clear();
 		// add all elements of the top chain	
 		std::vector< std::pair<unsigned,unsigned> > & V=vindex[heapbeg->v];
 		it=V.begin();
@@ -2512,8 +2688,9 @@ namespace giac {
 		  type_operator_plus_times_reduce((itbbeg+it->first)->g,(qbeg+its)->g,g,reduce);
 		  // increment 2nd poly index of the elements of the top chain
 		  ++its;
-		  if (its<qsize)
+		  if (its<qsize){
 		    nouveau.push_back(*it);
+		  }
 		  else // wait for computation of a new term of a before adding to the heap 
 		    qnouveau.push_back(*it);
 		}
@@ -2525,12 +2702,14 @@ namespace giac {
 #endif
 		// std::pop_heap(heapbeg,heapend);
 		--heapend;
+	      } // while heapend!=heapbeg && heapu==
+	      {
 		// push each element of the incremented top chain 
 		it=nouveau.begin();
 		itend=nouveau.end();
 		for (;it!=itend;++it) {
 		  u=(itbbeg+it->first)->u+(qbeg+it->second)->u;
-		  // if (norem && u<bu) continue;
+		  if (norem && one_index_smaller(u,bu,varsshift)) continue;
 		  // check if u is in the path to the root of the heap
 		  unsigned holeindex=unsigned(heapend-heapbeg),parentindex;
 		  if (holeindex && u==heapbeg->u){
@@ -2583,8 +2762,12 @@ namespace giac {
 	      }
 	      else
 		g=-g;
-	      if (is_zero(g))
+	      if (is_zero(g)){
+#ifdef HEAP_STATS
+		++nullq;
+#endif
 		continue;
+	      }
 	    } // end if (heapu>=bu)
 	  } // end else ita->u>heapu
 	} // if heap non empty
@@ -2616,19 +2799,27 @@ namespace giac {
 	std::vector< std::pair<unsigned,unsigned> >::iterator it,itend;
 	it=qnouveau.begin();
 	itend=qnouveau.end();
+	U prevu=0; int previndex=-1;
 	for (;it!=itend;++it){
 	  if (!it->first) // leading term of b already taken in account
 	    continue;
 	  u=(itbbeg+it->first)->u+(q.begin()+it->second)->u;
-	  // if we compute quotient only check that u is smaller than bu 
-	  //if (norem && u<bu) continue;
+	  if (norem && one_index_smaller(u,bu,varsshift)) continue;
+	  if (u==prevu && previndex>=0){
+#ifdef HEAP_STATS
+	    ++chain;
+#endif
+	    vindex[previndex].push_back(*it);
+	    continue;
+	  }
+	  prevu=u;
 	  // check if u is in the path to the root of the heap
 	  unsigned holeindex=unsigned(heapend-heapbeg),parentindex;
 	  if (holeindex && u==heapbeg->u){
 #ifdef HEAP_STATS
 	    ++chain;
 #endif
-	    vindex[heapbeg->v].push_back(*it);
+	    vindex[previndex=heapbeg->v].push_back(*it);
 	    continue;
 	  }
 	  bool done=false;
@@ -2642,7 +2833,7 @@ namespace giac {
 	      ++chain;
 #endif
 	      done=true;
-	      vindex[(heapbeg+parentindex)->v].push_back(*it);
+	      vindex[previndex=(heapbeg+parentindex)->v].push_back(*it);
 	      break;
 	    }
 	    holeindex=parentindex;
@@ -2652,7 +2843,7 @@ namespace giac {
 	    ++nochain;
 #endif
 	    // not found, add a new node to the heap
-	    std::vector< std::pair<unsigned,unsigned> > & V=vindex[heapend->v];
+	    std::vector< std::pair<unsigned,unsigned> > & V=vindex[(previndex=heapend->v)];
 	    V.clear();
 	    V.push_back(*it);
 	    heapend->u=u;
@@ -2669,11 +2860,11 @@ namespace giac {
       } // for (;;)
 #ifdef HEAP_STATS
       if (debug_infolevel)
-	CERR << "chain " << chain << ", nochain " << nochain << ", type_op_reduce " << typeopreduce << std::endl;
+	CERR << "chain " << chain << ", nochain " << nochain << ", type_op_reduce " << typeopreduce << " null quotients" << nullq << std::endl;
 #endif
       // r still empty
       if (debug_infolevel>2)
-	CERR << "Finished computing quotient, size " << q.size() << " " << CLOCK() << std::endl ;
+	CERR << CLOCK()*1e-6 << " Finished computing quotient, size " << q.size() << std::endl ;
       if (quo_only==2 || quo_only==-2){
 	delete [] heap;
 	return 1;
@@ -2683,7 +2874,7 @@ namespace giac {
 	double qb=double(q.size())*b.size();
 	qb /= a.size();
 	if (debug_infolevel>1)
-	  CERR << CLOCK() << " qb=" << qb << std::endl;
+	  CERR << CLOCK()*1e-6 << " qb=" << qb << std::endl;
 	if (qb>100){
 	  // the coefficients might be not optimal (mpz_class instead of int)
 	  if (hashdivrem_finish_later(a.front().g)){
@@ -2704,15 +2895,15 @@ namespace giac {
 	  convert(bcopy,vars,newvars);
 	  convert(qcopy,vars,newvars);
 	  if (debug_infolevel>1)
-	    CERR << CLOCK() << " compress monomials done" <<std::endl;
-	  if (!threadmult(bcopy,qcopy,bq,vars.front(),reduce,a.size()))
+	    CERR << CLOCK()*1e-6 << " compress monomials done" <<std::endl;
+	  if (!threadmult(bcopy,qcopy,bq,newvars.front(),reduce,a.size()))
 	    smallmult(bcopy,qcopy,bq,reduce,as);
 	  smallsub(acopy,bq,r);
 	  if (debug_infolevel>1)
-	    CERR << CLOCK() << " uncompress monomials" <<std::endl;
+	    CERR << CLOCK()*1e-6 << " uncompress monomials" <<std::endl;
 	  convert(r,newvars,vars);
 	  if (debug_infolevel>1)
-	    CERR << CLOCK() << " uncompress monomials end"<< std::endl;
+	    CERR << CLOCK()*1e-6 << " uncompress monomials end"<< std::endl;
 	  delete [] heap;
 	  return 1;
 	}
@@ -2758,12 +2949,18 @@ namespace giac {
 	  // push each element of the incremented top chain 
 	  it=qnouveau.begin();
 	  itend=qnouveau.end();
+	  U prevu=0; int previndex=-1;
 	  for (;it!=itend;++it){
 	    u=(itbbeg+it->first)->u+(q.begin()+it->second)->u;
+	    if (u==prevu && previndex>=0){
+	      vindex[previndex].push_back(*it);
+	      continue;
+	    }
+	    prevu=u;
 	    // check if u is in the path to the root of the heap
 	    unsigned holeindex=unsigned(heapend-heapbeg),parentindex;
 	    if (holeindex && u==heapbeg->u){
-	      vindex[heapbeg->v].push_back(*it);
+	      vindex[(previndex=heapbeg->v)].push_back(*it);
 	      continue;
 	    }
 	    bool done=false;
@@ -2774,14 +2971,14 @@ namespace giac {
 		break;
 	      if (u==pu){
 		done=true;
-		vindex[(heapbeg+parentindex)->v].push_back(*it);
+		vindex[(previndex=(heapbeg+parentindex)->v)].push_back(*it);
 		break;
 	      }
 	      holeindex=parentindex;
 	    }
 	    if (!done){
 	      // not found, add a new node to the heap
-	      std::vector< std::pair<unsigned,unsigned> > & V=vindex[heapend->v];
+	      std::vector< std::pair<unsigned,unsigned> > & V=vindex[(previndex=heapend->v)];
 	      V.clear();
 	      V.push_back(*it);
 	      heapend->u=u;
